use std::collections::BTreeMap;
use std::error::Error;
use std::fmt;

use toml::Table;

use plugins::Plugin;
use plugins::hello::Hello;
use plugins::discord::DiscordPlugin;
use plugins::lua::LuaPlugin;
use plugins::stdin::StdinPlugin;
use plugins::echo::EchoPlugin;

pub struct PluginLoader {
    types: BTreeMap<&'static str, fn(&Table) -> Box<Plugin>>
}

impl PluginLoader {
    pub fn new () -> PluginLoader {
        let mut types = BTreeMap::new();
        types.insert("hello", Hello::new as fn(&Table) -> Box<Plugin>);
        types.insert("discord", DiscordPlugin::new as fn(&Table) -> Box<Plugin>);
        types.insert("lua", LuaPlugin::new as fn(&Table) -> Box<Plugin>);
        types.insert("stdin", StdinPlugin::new as fn(&Table) -> Box<Plugin>);
        types.insert("echo", EchoPlugin::new as fn(&Table) -> Box<Plugin>);
        PluginLoader {
            types: types
        }
    }

    pub fn load_from_configuration (&self, configuration: Table) -> Result<BTreeMap<String, Box<Plugin>>, PluginLoaderError> {
        configuration.into_iter().map(|(key, value)| {
            match value.as_table() {
                Some(table) => {
                    let plugin = self.load_single_plugin(&key, table)?;
                    Result::Ok((key, plugin))
                },
                None => Result::Err(PluginLoaderError { message: format!("Bad configuration parameters for plugin instance: {}. Configuration for a plugin must be a table.", key) })
            }
        }).collect()
    }

    pub fn load_single_plugin (&self, name: &str, configuration: &Table) -> Result<Box<Plugin>, PluginLoaderError> {
        /*
         * The plugin type defaults to the instance name (in the tenquestionmarks configuration)
         * but can explicitly be set by using the special "type" parameter.
         */
        let plugin_type: &str = configuration.get("type")
                                             .and_then(|value| value.as_str())
                                             .unwrap_or(name);

        match self.types.get(plugin_type) {
           Some(constructor) => Result::Ok(constructor(configuration)),
           None => Result::Err(PluginLoaderError { message: format!("No such plugin type: {}", plugin_type) })
        }
    }
}

#[derive(Debug)]
pub struct PluginLoaderError {
    message: String
}

impl Error for PluginLoaderError {
    fn description(&self) -> &str {
        &self.message[..]
    }
}

impl fmt::Display for PluginLoaderError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "PluginLoaderError: {}", self.message)
    }
}
